#include <iostream>
#include <math.h>
#include <vector>

#include "hdnum.hh"

#include "model_arenstorf.hh"

template <typename S, typename M>
void solve(S solver, const M& model, bool plot=false)
{
  typedef typename S::number_type Number;
  Number time;
  hdnum::Vector<Number> x0(4,0);
  model.initialize(time,x0);

  std::vector<Number> times;
  std::vector<hdnum::Vector<Number> > states;
  if(plot){
    times.push_back(solver.get_time());
    states.push_back(solver.get_state());
  }

  // time loop
  Number T = 17.06521656015796255889172062;
  hdnum::Timer timer;
  timer.reset();
  while (solver.get_time()<T){
    // advance model and save time and state
    solver.step();
    if(plot){
      times.push_back(solver.get_time());
      states.push_back(solver.get_state());
    }

    // last time step to T
    Number dtlocal = solver.get_dt();
    
    if ((solver.get_time()+dtlocal) > T)
      solver.set_dt(T-solver.get_time());
  
  }
  double compTime = timer.elapsed();

  // error at time T
  Number xx = solver.get_state()[0]-x0[0];
  Number yy = solver.get_state()[1]-x0[1];
  std::cout << "Error: "
            << std::scientific << sqrt(xx*xx+yy*yy) << std::endl;

  // f evaluations
  std::cout << "f evaluations: "
            << model.f_evaluations() << std::endl;

  // print elapsed time
  std::cout << "Computation time: "
            << compTime << std::endl;

  // save results for plotting
  if(plot){
    gnuplot("threebody.dat",times,states);
  }

}

int main(int argc, char ** argv)
{
  if(argc < 5){
    // note: with that many parameters you should start thinking about using
    // a config file parser ;)
    std::cout << "Usage  : ./uebung4 solver dt TOL plot(bool)" << std::endl;
    std::cout << "possible solvers:" << std::endl;
    std::cout << "1: RungeKutta4" << std::endl;
    std::cout << "2: RKF45" << std::endl;
    std::cout << std::endl;
    std::cout << "if solver doesn't need dt or TOL just give any number"
              << std::endl;
    exit(0);
  }

  typedef double Number;

  // read command line parameters
  int solverType(atoi(argv[1]));
  Number dt(atof(argv[2]));
  Number TOL(atof(argv[3]));
  bool plot(atoi(argv[4]));

  // define Problem
  typedef ArenstorfOrbitModel<Number> Model;
  Model model;


  // solve problem
  if(solverType==1){
    typedef hdnum::RungeKutta4<Model> Solver;
    Solver solver(model);
    solver.set_dt(dt);
    std::cout << "=== solve with RungeKutta4" << std::endl;
    solve(solver,model,plot);
  }

  if(solverType==4){
    typedef hdnum::RKF45<Model> Solver;
    Solver solver(model);
    solver.set_TOL(TOL);
    std::cout << "=== solve with RKF45" << std::endl;
    solve(solver,model,plot);
  }

  return 0;
}
